#include <fstream>
#include <cstdio>
#include <vector>
#include <string>
#include <argh.hh>

using namespace std;

#define FOURCC 0x52474218  // BGR24 fourcc

static FILE *fout;
static unsigned frameus = 1000000 / 25;
static vector<string> files;

static unsigned width, height;

static bool xflip = false;
static bool yflip = false;

static void dwf(unsigned dw)
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
    fputc((unsigned char)((dw >> 16) & 255), fout);
    fputc((unsigned char)((dw >> 24) & 255), fout);
}
static void wf(unsigned short dw)
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
}

class SizeHere
{
    unsigned pos;
public:
    SizeHere()
    {
        pos = ftell(fout);
        fwrite("----", 4, 1, fout);
    }
    ~SizeHere()
    {
        unsigned now = ftell(fout);
        fseek(fout, pos, SEEK_SET);
        dwf(now - pos - 4);
        fseek(fout, now, SEEK_SET);
    }
};

static bool isValidTGA(FILE* fp)
{
    // Check that it's an uncompressed 24-bit tga:
    fseek(fp, 2, SEEK_SET);
    int tgaType = fgetc(fp);
    fseek(fp, 16, SEEK_SET);
    int bitsPerPixel = fgetc(fp);
    if(tgaType != 2 || bitsPerPixel != 24) return false;
    return true;
}

static void CalcSize()
{
    width = 0;
    height = 0;
    for(unsigned i=0; i<files.size(); ++i)
    {
        FILE *fp = fopen(files[i].c_str(), "rb");
        if(fp && isValidTGA(fp))
        {
            fseek(fp, 12, SEEK_SET);
            width = fgetc(fp);  width |= fgetc(fp) << 8;
            height = fgetc(fp); height |= fgetc(fp) << 8;
            fclose(fp);
            return;
        }
        fclose(fp);
    }
}

static void WriteFile(const string &fn)
{
    FILE *fp = fopen(fn.c_str(), "rb");
    if(!fp)
    {
        perror(fn.c_str());
        return;
    }

    unsigned datapos = fgetc(fp) + 18;

    if(!isValidTGA(fp))
    {
        fprintf(stderr, "%s is not an uncompressed 24-bit tga.\n", fn.c_str());
        return;
    }

    // Check that image dimensions are the same as in first tga:
    fseek(fp, 12, SEEK_SET);
    unsigned new_width = fgetc(fp);  new_width |= fgetc(fp) << 8;
    unsigned new_height = fgetc(fp); new_height |= fgetc(fp) << 8;
    if(new_width != width || new_height != height)
    {
        fprintf(stderr, "%s (%ux%u) had different size than the first "
                "valid image (%ux%u)\n", fn.c_str(),
                new_width, new_height, width, height);
        return;
    }
    
    char *Buf = new char[width*3];
    for(unsigned y=0; y<height; ++y)
    {
        unsigned yy = yflip ? (height - y - 1) : y;
        
        fseek(fp, datapos + yy*width*3, SEEK_SET);
        fread(Buf, 3, width, fp);
        
        if(xflip)
            for(unsigned x=width; x-->0; )fwrite(&Buf[x*3], 1, 3, fout);
        else
            fwrite(Buf, 3, width, fout);
    }
    delete[] Buf;
    
    fclose(fp);
}
static void WriteSTRH()
{
    fwrite("strh", 4, 1, fout);
    SizeHere tmp;
    fwrite("vids", 4, 1, fout);
    dwf(FOURCC);// FOURCC
    dwf(0);     // flags
    wf(0);      // priority
    wf(0);      // language
    dwf(0);     // initial frames
    dwf(1000000);                  // scale
    dwf((unsigned)(1e12/frameus)); // rate  (put scale*fps here)
    dwf(0);     // start
    dwf(0);     // scale*framecount
    dwf(1048576);// suggested buffer size
    dwf(10000); // quality
    dwf(width*height*3);
    wf(0);       // left
    wf(0);       // top
    wf(width);   // right
    wf(height);  // bottom
}
static void WriteSTRF()
{
    fwrite("strf", 4, 1, fout);
    SizeHere tmp;
    dwf(40);    // size - ????
    dwf(width); // width
    dwf(height);// height
    wf(1);      // planes
    wf(24);     // depth
    /* compression type */
    dwf(FOURCC);     // FOURCC
    dwf(width*height*3); // width*height*3
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
}
static void WriteAVIH()
{
    fwrite("avih", 4, 1, fout);
    SizeHere tmp;
    dwf(frameus);   // microseconds per frame
    dwf((unsigned)(width*height*3e6/frameus)); // max bytes per sec
    dwf(0);         // padding (0)
    dwf(0);         // flags
    dwf(files.size());// total frames
    dwf(0);         // initial frames
    dwf(1);         // streamcount (1, only one: video)
    dwf(1048576);   // suggested buffer size
    dwf(width);     // width
    dwf(height);    // height
    dwf(0);dwf(0);dwf(0);dwf(0); // reserved
}
static void WriteSTRL()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("strl", 4, 1, fout);
    WriteSTRH();
    WriteSTRF();
}
static void WriteHDRL()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("hdrl", 4, 1, fout);
    WriteAVIH();
    WriteSTRL();
}
static void WriteMOVI()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("movi", 4, 1, fout);
    
    for(unsigned a=0; a<files.size(); ++a)
    {
        char Buf[5] = "00dc"; // 00 = stream index
        fwrite(Buf, 4, 1, fout);
        SizeHere tmp;
        WriteFile(files[a]);
    }
}

static void WriteAVI()
{
    fwrite("RIFF", 4, 1, fout);
    SizeHere tmp;
    fwrite("AVI ", 4, 1, fout);
    WriteHDRL();
    WriteMOVI();
}

bool readInputFilenames(const string& listfile)
{
    ifstream is(listfile.c_str());
    if(!is)
    {
        fprintf(stderr, "Can't open "); perror(listfile.c_str());
        return false;
    }

    string filename;

    while(is)
    {
        getline(is, filename);
        if(is.good() && filename.size() && filename[0]!='#')
        {
            while(filename.size() && filename[0]==' ') filename.erase(0,1);
            while(filename.size() && filename[filename.size()-1]==' ')
                filename.erase(filename.size()-1,1);
            if(filename.size())
                files.push_back(filename);
        }
    }

    return true;
}

int main(int argc, const char *const *argv)
{
    string outfile;
    
    ParamHandler Argh;
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("outfile", 'o').SetString(1).SetDesc("Specify the output file.", "<filename>");
    Argh.AddLong("fps",     'f').SetFloat(23e-5, 1e6).SetDesc("Frame rate (default: 25).", "<float>");
    Argh.AddLong("xflip",   'X').SetBool().SetDesc("Flip horizontally (not default)");
    Argh.AddLong("yflip",   'Y').SetBool().SetDesc("Flip vertically (default)");
    Argh.AddLong("list",    '@').SetString(1).SetDesc("Read input file names from the given file.", "<filename>");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is animmaker-"VERSION"\n"
                    "Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: animmaker [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software assumes the files are uncompressed rgb24 tga files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                return 0;
            case 'o':
                outfile = Argh.GetString();
                break;
            case 'X':
                xflip = Argh.GetBool();
                break;
            case 'Y':
                yflip = Argh.GetBool();
                break;
            case 'f':
                frameus = (unsigned)(1e6 / Argh.GetFloat());
                break;
            case '@':
                if(!readInputFilenames(Argh.GetString())) return -1;
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified. Try --help.\n");
        return -1;
    }
    
    if(!outfile.size() || outfile == "-")
    {
        fout = stdout;
    }
    else
        fout = fopen(outfile.c_str(), "wb");
    if(!fout)
    {
        perror(outfile.c_str());
        return -1;
    }
    
    CalcSize();
    WriteAVI();
    
    fclose(fout);
    
    return 0;
}
